#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Helper for building image ingestion manifest """
import logging
from pathlib import Path
from typing import List

from ingest_envoy.manifest_components import (
    TARFILE_EXT,
    WEBLOG_FILENAME,
    AncillaryProduct,
    OutputScienceProduct,
)
from ingest_envoy.utilities import AncillaryProductType

FITS = "fits"
RMS = "rms"
PB = "pb"
MASK = "mask"


# pylint: disable=R1721
class ImageIngestionProductsFinder:
    """Finds ancillary science products and other ancillary products needed for image ingestion"""

    def __init__(self, staging_source_dir: Path):
        self.logger = logging.getLogger("ingest_envoy")
        self.staging_source_dir = staging_source_dir
        self.files_found = [file for file in self.staging_source_dir.iterdir()]
        self.output_science_products = self._find_output_science_products()
        self.ancillary_products = self._find_other_ancillary_products()

    def _find_output_science_products(self) -> List[OutputScienceProduct]:
        """
        Find the ancillary products belonging to the science product*
        in the staging dir.

        N.B. ingestion artifacts file will not be there until manifest has been created!

        * in our example there is exactly one science product with a couple of ancillary
        products, so for now the dict we return will contain only one science product.
        if this changes for some new type of science product type, we'll have to figure
        out how to determine which ancillary product(s) belong to which science product.


        :return:
        """
        sp_image_file = [
            file
            for file in self.files_found
            if file.name.endswith(FITS) and RMS not in file.name and PB not in file.name and MASK not in file.name
        ][0]
        self.logger.info(f"Science Product to ingest is: {sp_image_file.name}")
        image_files = [
            file
            for file in self.files_found
            if (file.name.endswith(FITS) or file.name.endswith(".png")) and file.name != sp_image_file.name
        ]
        self.logger.info(f"Science Product ancillaries are: {image_files}")

        sp_aps = []
        for image_file in image_files:
            sp_ap = self._build_ancillary_image_science_product(image_file)
            sp_aps.append(sp_ap)

        # TODO: what type of image is this science product actually?
        #  It's a FITS in our example, but this may not always be the case
        science_product = OutputScienceProduct(
            type=AncillaryProductType.FITS,
            filename=sp_image_file.name,
            ancillary_products=sp_aps,
        )

        return [science_product]

    def _find_other_ancillary_products(self) -> List[AncillaryProduct]:
        """
        Find the "other" ancillary image products in the staging dir: there should be a
        weblog and a pipeline artifacts tar. (The ingestion artifacts tar will be produced
        during the building of the manifest.)

        :return:
        """

        try:
            pipeline_artifacts_tar = [
                file
                for file in self.files_found
                if file.name.startswith(AncillaryProductType.PIPELINE_ARTIFACTS.value)
                and file.name.endswith(TARFILE_EXT)
            ][0]
        except IndexError as err:
            raise FileNotFoundError(f"WARNING: No pipeline artifacts found in {self.staging_source_dir}") from err

        ancillary_products = [
            AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=WEBLOG_FILENAME),
            AncillaryProduct(type=AncillaryProductType.PIPELINE_ARTIFACTS, filename=pipeline_artifacts_tar.name),
        ]

        return ancillary_products

    def _build_ancillary_image_science_product(self, file: Path):
        """
        Image science products will have ancillary products of their very own,
        distinct from other ancillary products that might be in the staging dir.

        :param file: a possible ancillary image product
        :return: the corresponding AncillaryProduct, if applicable
        """
        filename = file.name
        if "image" in filename:
            if filename.endswith(".png"):
                return AncillaryProduct(type=AncillaryProductType.THUMBNAIL_IMG, filename=filename)
            if filename.endswith(FITS) and RMS in filename:
                return AncillaryProduct(type=AncillaryProductType.QUICKLOOK_RMS_IMAGE, filename=filename)
            return AncillaryProduct(type=AncillaryProductType.QUICKLOOK_IMAGE, filename=filename)
        if filename.endswith(FITS):
            if PB in filename:
                return AncillaryProduct(type=AncillaryProductType.PB_FITS, filename=filename)
            if MASK in filename:
                return AncillaryProduct(type=AncillaryProductType.FITS_MASK, filename=filename)

        # TODO: what might it be?
        raise ValueError(f"don't know what type of image '{filename}' is")

    def _get_product_root(self):
        """
        There should be a common anchor for the science products and ancillary products
        of an ingestion manifest. In our example, it's
        VLASS2.1.ql.T08t09.J055438_113000.10.2048.v1.

        :return:
        """
        files = [file for file in self.staging_source_dir.iterdir()]
        for file in files:
            # Okay, this is kludgy, but it works with our example manifest
            # and currently that's all we have to go on. -- JLG, 2021-07-30
            if file.name.endswith(".fits"):
                parts = file.name.split(".v")
                return parts[0]
        raise ValueError(
            f"Unable to determine the product root from the output science products" f" in {self.staging_source_dir}"
        )

    def _is_ancillary_image_product(self, file: Path) -> bool:
        """
        Is this the kind of ancillary product that will go in the "ancillary_products"
        section of the manifest (and -not- an ancillary product belonging to
        a science product?

        :param file: some file found in the ingestion staging dir
        :return: whether it's an ancillary product
        """
        if file.name == WEBLOG_FILENAME or "ingestion_artifacts" in file.name:
            return True
        product_root = self._get_product_root()
        return file.name.startswith(product_root) and file.name.endswith(TARFILE_EXT)
