#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""Regression & debugging tests for CapabilityRequestStateSetter"""

from workspaces.capability.enums import CapabilityRequestState, CapabilityVersionState
from workspaces.capability.schema import (
    CapabilityRequest,
    CapabilityRequestStateSetter,
    CapabilityVersion,
)


def test_works_for_one_created():
    """
    Given: a capability request with one version
    And: that version is in the Created state
    When: CapabilityRequestStateSetter determines request state
    Then: request state is Created

    :return:
    """
    request = CapabilityRequest()
    request.id = 1
    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Created.name, request=request, capability_request_id=request.id
    )
    CapabilityRequestStateSetter(request).determine_state()
    assert request.state == CapabilityRequestState.Created.name


def test_all_versions_same_still_works():
    """
    Given: a capability request with several versions all in same state
    When: CapabilityRequestStateSetter determines request state
    Then: request state corresponds to each version's state

    :return:
    """
    for state in CapabilityVersionState:
        request = CapabilityRequest()
        request.id = 1
        versions = []
        for i in range(4):
            version = CapabilityVersion()
            version.version_number = i + 1
            version.state = state.name
            version.request = request
            version.capability_request_id = request.id
            versions.append(version)
        CapabilityRequestStateSetter(request).determine_state()
        for version in request.versions:
            assert version_state_matches_request_state(version, request)


def test_any_version_complete_still_works():
    """
    Given: a capability request with several versions
    When: CapabilityRequestStateSetter determines request state
    Then: correct request state is set depending on whether or not at least one of its version is Complete

    :return:
    """
    request = CapabilityRequest(id=1)
    assert not CapabilityRequestStateSetter(request).any_version_complete()

    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Created.name, request=request, capability_request_id=request.id
    )
    setter = CapabilityRequestStateSetter(request)
    assert not setter.any_version_complete()
    setter.determine_state()
    assert request.state == CapabilityRequestState.Created.name

    request = CapabilityRequest(id=1)
    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Complete.name, request=request, capability_request_id=request.id
    )
    assert CapabilityRequestStateSetter(request).any_version_complete()

    request = CapabilityRequest(id=1)
    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Created.name, request=request, capability_request_id=request.id
    )
    CapabilityVersion(
        version_number=2, state=CapabilityVersionState.Complete.name, request=request, capability_request_id=request.id
    )
    setter = CapabilityRequestStateSetter(request)
    assert setter.any_version_complete()

    request = CapabilityRequest(id=1)
    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Created.name, request=request, capability_request_id=request.id
    )
    CapabilityVersion(
        version_number=2, state=CapabilityVersionState.Running.name, request=request, capability_request_id=request.id
    )
    CapabilityVersion(
        version_number=3, state=CapabilityVersionState.Failed.name, request=request, capability_request_id=request.id
    )
    setter = CapabilityRequestStateSetter(request)
    assert not setter.any_version_complete()
    setter.determine_state()
    assert request.state == CapabilityRequestState.Submitted.name

    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Created.name, request=request, capability_request_id=request.id
    )
    CapabilityVersion(
        version_number=2, state=CapabilityVersionState.Running.name, request=request, capability_request_id=request.id
    )
    CapabilityVersion(
        version_number=3, state=CapabilityVersionState.Complete.name, request=request, capability_request_id=request.id
    )
    setter = CapabilityRequestStateSetter(request)
    assert setter.any_version_complete()


def test_still_works_for_one_complete_all_others_failed():
    """
    Given: a capability request with several versions, one complete, the rest Failed
    When: CapabilityRequestStateSetter determines request state
    Then: request state is set to Complete

    :return:
    """
    request = CapabilityRequest(id=1)
    assert not CapabilityRequestStateSetter(request).any_complete_all_others_failed()

    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Complete.name, request=request, capability_request_id=request.id
    )
    assert CapabilityRequestStateSetter(request).any_complete_all_others_failed()
    CapabilityRequestStateSetter(request).determine_state()
    assert request.state == CapabilityRequestState.Complete.name

    request = CapabilityRequest(id=1)
    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Failed.name, request=request, capability_request_id=request.id
    )
    assert not CapabilityRequestStateSetter(request).any_complete_all_others_failed()

    request = CapabilityRequest(id=1)
    CapabilityVersion(
        version_number=1, state=CapabilityVersionState.Failed.name, request=request, capability_request_id=request.id
    )
    CapabilityVersion(
        version_number=2, state=CapabilityVersionState.Complete.name, request=request, capability_request_id=request.id
    )
    assert CapabilityRequestStateSetter(request).any_complete_all_others_failed()
    CapabilityVersion(
        version_number=3, state=CapabilityVersionState.Complete.name, request=request, capability_request_id=request.id
    )


def version_state_matches_request_state(version: CapabilityVersion, request: CapabilityRequest) -> bool:
    """
    Did CapabilityRequestSetter get it right?

    :param version:
    :param request:
    :return:
    """
    # a switch()! a switch()! my kingdom for a switch()!
    if version.state == CapabilityVersionState.Running.name:
        return request.state == CapabilityRequestState.Submitted.name
    if version.state == CapabilityRequestState.Created.name:
        return request.state == CapabilityRequestState.Created.name
    if version.state == CapabilityRequestState.Complete.name:
        return request.state == CapabilityRequestState.Complete.name
    if version.state == CapabilityRequestState.Failed.name:
        return request.state == CapabilityRequestState.Failed.name

    raise ValueError(f"Unrecognized capability version state: {version.state}")
